#pragma once
#include <iostream>
#include <future>
#include <memory>
#include <thread>
#include <boost/asio.hpp>
#include <boost/asio/spawn.hpp>

using tcp_socket = boost::asio::ip::tcp::socket;
using yield_context = boost::asio::yield_context;
using io_service = boost::asio::io_service;
using tcp_endpoint = boost::asio::ip::tcp::endpoint;
using tcp_resolver = boost::asio::ip::tcp::resolver;
using tcp_query = tcp_resolver::query;

template <typename endpoint_type>
struct my_connect_condition;

class TCPSession;

template <typename endpoint_type> my_connect_condition<endpoint_type> make_connect_condition(const endpoint_type &endpoint, TCPSession *client);

template<typename context_t, typename result_t, class class_t, typename ...func_args_t, typename ...args_t>
std::future<result_t> when(context_t &ctx, class_t *_this,
	result_t(class_t::*func)(boost::asio::yield_context&, func_args_t...), args_t ...args)
{
	std::packaged_task<result_t(boost::asio::yield_context)> task{
		std::bind(func, _this, std::placeholders::_1, std::forward<args_t>(args)...)
	};
	auto future = task.get_future();
	boost::asio::spawn(ctx, std::move(task));
	return future;
}

class TCPSession : public std::enable_shared_from_this<TCPSession>
{
protected:
	TCPSession(io_service &service);
	tcp_socket socket;
	std::string host_;
	tcp_endpoint endpoint_;

public:
	size_t operation_;

	void Connect(yield_context &yield, const char *host, unsigned short port);
	void Disconnect(yield_context &yield);

	std::future<void> Connect(const char *host, unsigned short port);
	std::future<void> Disconnect();
};

template <typename endpoint_type>
struct my_connect_condition
{
	TCPSession *client_;
	endpoint_type endpoint_;
	bool first_;
	size_t operation_; //copy of TCPClient operation_ int. If this changes this connection attempt was cancelled. 

	my_connect_condition(TCPSession *client,
		const endpoint_type &endpoint)
		: client_(client), endpoint_(endpoint),
		operation_(client->operation_), first_(true) {};

	template <typename Iterator>
	Iterator operator()(
		const boost::system::error_code& ec,
		Iterator next)
	{
		if (client_->operation_ != operation_)
		{
			std::cerr << "cancelled connecting to " << endpoint_ << std::endl;
			return Iterator{};
		}
		else if (ec)
		{
			std::cerr << "Error connecting to [" << endpoint_ << "]: " << ec.message() << std::endl;
		}
		else if (!first_)
		{
			std::cout << "Successfully connected to " << endpoint_ << std::endl;
			return Iterator{};
		}
		else
			first_ = false;

		endpoint_ = next->endpoint();
		std::cout << "Trying: " << endpoint_ << std::endl;
		return next;
	}
};

template <typename endpoint_type> my_connect_condition<endpoint_type> make_connect_condition(const endpoint_type &endpoint, TCPSession *client)
{
	return my_connect_condition<endpoint_type>(client, endpoint);
}


